% :- type code     == list(int).
% :- type blacks   == int.
% :- type whites   == int.
% :- type answer ---> blacks/whites.
% :- type hint   ---> code-answer.
% :- type file      == atom.

:- module(mmframe, [
        mmind_in/3,	% mmind_in(file::in, int::out, list(hint)::out) 
        mmind_out/2,    % mmind_out(file::in, list(code)::in)
        solve/2         % solve(file::in, file::in)
		   ]
	 ).

:- use_module(library(lists), [append/3]).
%-------------- Bejaratok ------------------------------------

% mmind_in(+File, -Colours, -Hints):
% A File-bol beolvashato egy Colours db. szint hasznalo, Hints sugasokat
% megado feladat.
mmind_in(File, Colours, Hints) :-
	open_file(File, read, Stream),
	read_in(Stream, Characters),
	close(Stream),
	parse_data(Characters, Colours, Hints).

% mmind_out(+File, +Solution):
% A Solution megoldast kiirja a File allomanyba.
mmind_out(File, Solution) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_int_line(Solution),
		     (set_output(OOut),close(Stream))).

% solve(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
solve(FileIn, FileOut) :-
	mmind_in(FileIn, Colours, Hints),
	ensure_loaded(user:mmind),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Colours, Hints),
		     (set_output(OOut),close(Stream))).

% print_solutions(+Colours, +Hints):
% A Colours db. szint hasznalo, Hints sugasokat megado feladat minden
% megoldasat kiirja az aktualis kimenetre.
print_solutions(Colours, Hints) :-
	user:mmind(Colours, Hints, Solution),
	print_int_line(Solution),
	fail.
print_solutions(_).

% print_int_line(+L):
% Egy-egy szokozzel elvalasztva kirja L elemeit es uj sort az aktualis
% kimenetre.
print_int_line([]) :-
	nl.
print_int_line([H|T]) :-
	write(H), tab(1),
	print_int_line(T).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Colours, -Hints):
% Chars fuzerbol kielemezheto egy Colours db. szint hasznalo Hints sugasokat
% megado feladat.
parse_data(Chars, Colours, Hints) :-
	first_int(Colours, Chars, Chars1),
	int_lists(IntLists, Chars1, Rest),
	correct_syntax(Rest),
	int_lists_hints(IntLists, Hints).

% int_lists_hints(+IL, -HL):
% Az IL egeszek listajanak listaja a HL sugasok listajat adja.
int_lists_hints([], []).
int_lists_hints([HL|TL], [Code-Blacks/Whites|TH]) :-
	append(Code, [Blacks,Whites], HL),
	int_lists_hints(TL, TH).

%-------------- Ellenorzesek ---------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :- 
	Rest = [_|_], !, 
	chars_of_type(mid_line, L, Rest, _),
	warning('non-digit character in line: "~s"', [L]),
	fail.
correct_syntax(_).

%-------------- DCG elemzes ----------------------------------

% first_int(-Int) -->
% Kielemezheto az Int egesz, amit esetleg nem lathato karakterek
% eloznek meg.
first_int(Int) -->
	chars_of_type(layout, _),
	int_number(Int).

% int_lists(-LL) -->
% Kielemezheto az LL egeszek listainak listaja.
int_lists([IntList|IntLists]) -->
	chars_of_type(layout, _),
	int_list(IntList), !, 
	int_lists(IntLists).
int_lists([]) --> [].

% int_list(-IntList) -->
% Kielemezheto az IntList egeszek listaja, ujsorral lezarva.
% Az egeszeket vizszintes ures karakterek valasztjak el es egy
% `/' karakter is elofordulhat kozottuk.
% A bemenet nem kezdodhet nem lathato karakterrel.
int_list([Int|IntList]) -->
	int_number(Int), !,
	chars_of_type(horiz_sp, _),
	opt_slash,
	chars_of_type(horiz_sp, _),
	int_list(IntList).
int_list([]) -->
	char_of_type(vert_sp, _).

% int_number(-Int) -->
% Kielemezheto egy (tizes szamrendszerbeli) Int nem-negativ egesz szam.
int_number(Int) -->
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% opt_slash -->
% Esetleg kielemezheto egy `/' karakter.
opt_slash -->
	[0'/], !.
opt_slash --> [].

% layout_char(Dir, C):
% C egy Dir (horiz_sp|vert_sp) iranyu nem lathato karakter.
layout_char(horiz_sp, 0' ).
layout_char(horiz_sp, 0'\t).
layout_char(vert_sp,  0'\n).
layout_char(vert_sp,  0'\r).
layout_char(vert_sp,  0'\f).
layout_char(vert_sp,  0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	layout_char(_, C).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ layout_char(vert_sp, C).
char_type(Dir, C) :-
	layout_char(Dir, C).

